# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from falconswagger.exceptions import ModelBaseError
from myreco.items_types.models import build_item_key
from importlib import import_module


class ModuleClassLoader(object):
    _classes = dict()

    @classmethod
    def load(cls, config):
        key = '{}.{}'.format(config['path'], config['class_name'])
        class_ = cls._classes.get(key)

        if class_ is None:
            try:
                module = import_module(config['path'])
                class_ = getattr(module, config['class_name'])
                cls._classes[key] = class_

            except Exception as error:
                raise ModelBaseError("invalid module '{}.{}' configuration for this engine",
                                     config['path'], config['class_name'])

        return class_


def get_items_model_from_api(api, engine):
    if api:
        items_types_model_key = build_item_key(engine['item_type']['name'])
        items_collection = api.models.get(items_types_model_key)
        if items_collection:
            return items_collection.__models__.get(engine['store_id'])



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from falconswagger.utils import get_model_schema
from falconswagger.json_builder import JsonBuilder
from falconswagger.exceptions import ModelBaseError
from myreco.engines.cores.filters.factory import FiltersFactory
from myreco.utils import ModuleClassLoader, get_items_model_from_api
from falcon.errors import HTTPNotFound
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
import random as random_
import sqlalchemy as sa
import hashlib
import json


class PlacementsModelBase(AbstractConcreteBase):
    __tablename__ = 'placements'
    __schema__ = get_model_schema(__file__)

    hash = sa.Column(sa.String(255), unique=True, nullable=False)
    small_hash = sa.Column(sa.String(255), primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)
    ab_testing = sa.Column(sa.Boolean, default=False)
    show_details = sa.Column(sa.Boolean, default=True)
    distribute_recos = sa.Column(sa.Boolean, default=False)

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def variations(cls):
        return sa.orm.relationship('VariationsModel', uselist=True, passive_deletes=True)

    def __init__(self, session, input_=None, **kwargs):
        super().__init__(session, input_=input_, **kwargs)
        self._set_hash()

    def _set_hash(self):
        if self.name and not self.hash:
            hash_ = hashlib.new('ripemd160')
            hash_.update(self.name.encode() + bytes(self.store_id))
            self.hash = hash_.hexdigest()

    def __setattr__(self, name, value):
        if name == 'hash':
            self.small_hash = value[:5]

        super().__setattr__(name, value)

    @classmethod
    def get_recommendations(cls, req, resp):
        cls._get_recommendations(req, resp)

    @classmethod
    def get_slots(cls, req, resp):
        cls._get_recommendations(req, resp, True)

    @classmethod
    def _get_recommendations(cls, req, resp, by_slots=False):
        placement = cls._get_placement(req, resp)
        session = req.context['session']
        input_variables = req.context['parameters']['query_string']
        show_details = placement.get('show_details')
        distribute_recos = placement.get('distribute_recos')
        recos = []
        recos_key = 'slots' if by_slots else 'recommendations'

        for slot in placement['variations'][0]['slots']:
            slot_recos = {'fallbacks': []}
            slot_recos['main'] = \
                cls._get_recos_by_slot(slot, input_variables, session, show_details)

            cls._get_fallbacks_recos(slot_recos, slot, input_variables, session, show_details)

            if by_slots:
                slot = {'name': slot['name'], 'item_type': slot['engine']['item_type']['name']}
                slot['recommendations'] = slot_recos
                recos.append(slot)
            else:
                cls._get_recos(recos, slot_recos, slot, distribute_recos)

        if not recos:
            raise HTTPNotFound()

        if not by_slots:
            if distribute_recos:
                recos = cls._distribute_recos(recos)

            recos = cls._unique_recos(recos)

        placement = {'name': placement['name'], 'small_hash': placement['small_hash']}
        placement[recos_key] = recos

        resp.body = json.dumps(placement)

    @classmethod
    def _get_placement(cls, req, resp):
        small_hash = req.context['parameters']['path']['small_hash']
        session = req.context['session']
        placements = cls.get(session, {'small_hash': small_hash})

        if not placements:
            raise HTTPNotFound()

        return placements[0]

    @classmethod
    def _get_recos_by_slot(cls, slot, input_variables, session, show_details, max_recos=None):
        engine = slot['engine']
        items_model = get_items_model_from_api(cls.__api__, engine)
        engine_vars, filters = \
            cls._get_variables_and_filters(slot, items_model, input_variables)
        core_config = engine['core']['configuration']['core_module']
        core_instance = ModuleClassLoader.load(core_config)(engine, items_model)
        max_recos = slot['max_recos'] if max_recos is None else max_recos

        return core_instance.get_recommendations(
            session, filters, max_recos, show_details, **engine_vars)

    @classmethod
    def _get_variables_and_filters(cls, slot, items_model, input_variables):
        engine_vars = dict()
        filters = dict()
        engine = slot['engine']

        for engine_var in slot['slot_variables']:
            var_name = engine_var['variable']['name']
            var_engine_name = engine_var['inside_engine_name']

            if var_name in input_variables:
                schema, filter_input_schema = cls._get_variable_schema(engine, engine_var)

                filter_input_schema = schema if filter_input_schema is None else filter_input_schema
                var_value = JsonBuilder.build(input_variables[var_name], filter_input_schema)

                if not engine_var['is_filter']:
                    engine_vars[var_engine_name] = var_value
                else:
                    filter_ = FiltersFactory.make(items_model, engine_var, schema)
                    filters[filter_] = var_value

        return engine_vars, filters

    @classmethod
    def _get_variable_schema(cls, engine, engine_var):
        filter_input_schema = None

        if engine_var['is_filter']:
            if engine_var['filter_type'].endswith('Of'):
                filter_input_schema = {'type': 'array', 'items': engine['item_type']['schema']}

            variables = engine['item_type']['available_filters']
        else:
            variables = engine['variables']

        for var in variables:
            if var['name'] == engine_var['inside_engine_name']:
                if engine_var['is_filter'] and var['schema'].get('type') != 'array' \
                        and not engine_var['filter_type'].endswith('Of'):
                    filter_input_schema = {'type': 'array', 'items': var['schema']}

                return var['schema'], filter_input_schema

    @classmethod
    def _get_fallbacks_recos(cls, slot_recos, slot, input_variables, session, show_details):
        if len(slot_recos['main']) != slot['max_recos']:
            for fallback in slot['fallbacks']:
                fallbacks_recos_size = \
                    sum([len(fallback) for fallback in slot_recos['fallbacks']])
                max_recos = slot['max_recos'] - len(slot_recos['main']) - fallbacks_recos_size
                if max_recos == 0:
                    break

                fallback_recos = cls._get_recos_by_slot(
                    fallback, input_variables, session, show_details, max_recos)
                slot_recos['fallbacks'].append(fallback_recos)

    @classmethod
    def _get_recos(cls, recos, slot_recos, slot, distribute_recos):
        fallbacks_recos = []
        [fallbacks_recos.extend(recos) for recos in slot_recos['fallbacks']]
        slot_recos = slot_recos['main'] + fallbacks_recos

        for reco in slot_recos:
            reco['type'] = slot['engine']['item_type']['name']

        if distribute_recos:
            recos.append(slot_recos)
        else:
            recos.extend(slot_recos)

    @classmethod
    def _distribute_recos(cls, recos_list, random=True):
        total_length = sum([len(recos) for recos in recos_list])
        total_items = []
        initial_pos = 0

        for recos in recos_list:
            if not recos:
                continue

            step = int(total_length / len(recos))

            if random:
                initial_pos = random_.randint(0, step-1)

            positions = range(initial_pos, total_length, step)
            zip_items_positions = zip(recos, positions)
            total_items.extend(zip_items_positions)

        random_.shuffle(total_items)
        sorted_items = sorted(total_items, key=(lambda each: each[1]))

        return [i[0] for i in sorted_items]

    @classmethod
    def _unique_recos(cls, recos):
        unique = list()
        [unique.append(reco) for reco in recos if not unique.count(reco)]
        return unique

    @classmethod
    def _build_id_names_schema(cls, engine):
        return {
            'type': 'array',
            'items': [engine['item_type']['schema']['properties'][id_name] \
                for id_name in engine['item_type']['schema']['id_names']]
        }


class VariationsModelBase(AbstractConcreteBase):
    __tablename__ = 'variations'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)
    weight = sa.Column(sa.Float)

    @declared_attr
    def placement_hash(cls):
        return sa.Column(sa.ForeignKey('placements.hash', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    @declared_attr
    def slots(cls):
        return sa.orm.relationship('SlotsModel',
                uselist=True, secondary='variations_slots')


class ABTestUsersModelBase(AbstractConcreteBase):
    __tablename__ = 'ab_test_users'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)

    @declared_attr
    def variation_id(cls):
        return sa.Column(sa.ForeignKey('variations.id'), nullable=False)


def build_variations_slots_table(metadata, **kwargs):
    return sa.Table(
        'variations_slots', metadata,
        sa.Column('variation_id', sa.Integer, sa.ForeignKey('variations.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('slot_id', sa.Integer, sa.ForeignKey('slots.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.items_indices_map import ItemsIndicesMap
from myreco.engines.cores.filters.factory import FiltersFactory
from myreco.engines.cores.filters.filters import BooleanFilterBy
from falconswagger.models.orm.redis import ModelRedisMeta, ModelRedisFactory
from falconswagger.models.orm.redis_base import ModelRedisBase
from falconswagger.utils import get_model_schema, get_dir_path
from falconswagger.exceptions import ModelBaseError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
from jsonschema import ValidationError, Draft4Validator
from copy import deepcopy
from falcon import HTTPNotFound
import sqlalchemy as sa
import json


def build_item_key(name, store_id=None):
    name = name.lower().replace(' ', '_')
    if store_id:
        return '{}_{}'.format(name, store_id)
    return name


class ItemsTypesModelBase(AbstractConcreteBase):
    __tablename__ = 'items_types'
    __schema__ = get_model_schema(__file__)
    __build_items_models__ = True
    __schema_dir__ = get_dir_path(__file__)

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    schema_json = sa.Column(sa.Text, nullable=False)

    @declared_attr
    def stores(cls):
        return sa.orm.relationship('StoresModel', uselist=True, secondary='items_types_stores')

    def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'schema':
            self._validate_input(value)
            value = json.dumps(value)
            attr_name = 'schema_json'

        super()._setattr(attr_name, value, session, input_)

    def _validate_input(self, schema):
        Draft4Validator.check_schema(schema)

        for id_name in schema['id_names']:
            if id_name not in schema.get('properties', {}):
                raise ValidationError(
                    "id_name '{}' was not found in schema properties".format(id_name),
                    instance=schema['id_names'], schema=schema)

    def _format_output_json(self, dict_inst, todict_schema):
        if todict_schema.get('schema') is not False:
            if 'schema_json' in dict_inst:
                dict_inst['schema'] = json.loads(dict_inst.pop('schema_json'))

                schema_properties = dict_inst['schema'].get('properties', {})
                schema_properties_names = sorted(schema_properties.keys())
                dict_inst['available_filters'] = \
                    [{'name': name, 'schema': schema_properties[name]} \
                        for name in schema_properties_names]

    @classmethod
    def associate_all_items(cls, session):
        items_types = cls.get(session)
        cls.associate_items(items_types)

    @classmethod
    def associate_items(cls, items_types):
        [cls._build_items_collections_model(item_type) for item_type in items_types]

    @classmethod
    def _build_items_collections_model(cls, item_type):
        if cls.__api__:
            name, schema = item_type['name'], item_type['schema']
            id_names = schema['id_names']
            class_name = cls._build_class_name(name)
            key = build_item_key(name)
            items_models = {store['id']: cls._build_items_model(item_type, store) \
                for store in item_type['stores']}
            attributes = {
                '__key__': key,
                '__schema__': cls._build_items_collections_schema(key, schema, id_names),
                '__models__': items_models,
                '__all_models__': cls.__all_models__,
                '__item_type__': item_type,
                '__authorizer__': cls.__authorizer__
            }
            items_collections = \
                cls._get_items_collections_metaclass()(class_name, (ModelRedisBase,), attributes)
            cls.__api__.associate_model(items_collections)

    @classmethod
    def _build_class_name(cls, *names):
        final_name = ''
        for name in names:
            name = name.split(' ')
            for in_name in name:
                final_name += in_name.capitalize()

        return final_name + 'Model'

    @classmethod
    def _build_items_model(cls, item_type, store):
        class_name = cls._build_class_name(item_type['name'], store['name'])
        key = build_item_key(item_type['name'], store['id'])
        id_names = item_type['schema']['id_names']
        items_model = \
            ModelRedisFactory.make(class_name, key, id_names, metaclass=ItemsModelBaseMeta)
        items_model.__item_type__ = item_type
        return items_model

    @classmethod
    def _build_items_collections_schema(cls, key, schema, id_names):
        base_uri = '/{}'.format(key)
        id_names_uri = base_uri + '/' + '/'.join(['{{{}}}'.format(id_name) for id_name in id_names])
        patch_schema = deepcopy(schema)
        required = patch_schema.get('required')
        if required:
            patch_schema['required'] = [req for req in required if req in id_names]
        properties = patch_schema.get('properties')
        if properties:
            properties['_operation'] = {'enum': ['delete', 'update']}

        swagger_schema = {
            base_uri: {
                'parameters': [{
                    'name': 'Authorization',
                    'in': 'header',
                    'required': True,
                    'type': 'string'
                },{
                    'name': 'store_id',
                    'in': 'query',
                    'required': True,
                    'type': 'integer'
                }],
                'post': {
                    'parameters': [{
                        'name': 'body',
                        'in': 'body',
                        'required': True,
                        'schema': {
                            'type': 'array',
                            'minItems': 1,
                            'items': schema
                        }
                    }],
                    'operationId': 'post_by_body',
                    'responses': {'201': {'description': 'Created'}}
                },
                'patch': {
                    'parameters': [{
                        'name': 'body',
                        'in': 'body',
                        'required': True,
                        'schema': {
                            'type': 'array',
                            'minItems': 1,
                            'items': patch_schema
                        }
                    }],
                    'operationId': 'patch_by_body',
                    'responses': {'200': {'description': 'Updated'}}
                },
                'get': {
                    'parameters': [{
                        'name': 'page',
                        'in': 'query',
                        'type': 'integer',
                        'default': 1
                    },{
                        'name': 'items_per_page',
                        'in': 'query',
                        'type': 'integer',
                        'default': 1000
                    }],
                    'operationId': 'get_by_body',
                    'responses': {'200': {'description': 'Got'}}
                },
            },
            id_names_uri: {
                'parameters': [{
                    'name': 'Authorization',
                    'in': 'header',
                    'required': True,
                    'type': 'string'
                },{
                    'name': 'store_id',
                    'in': 'query',
                    'required': True,
                    'type': 'integer'
                }],
                'get': {
                    'operationId': 'get_by_uri_template',
                    'responses': {'200': {'description': 'Got'}}
                }
            }
        }

        base_uri_get_parameters = cls._build_id_names_parameters_schema(id_names, schema, 'query')
        swagger_schema[base_uri]['get']['parameters'].extend(base_uri_get_parameters)

        id_names_uri_parameters = cls._build_id_names_parameters_schema(id_names, schema, 'path')
        swagger_schema[id_names_uri]['parameters'].extend(id_names_uri_parameters)

        return swagger_schema

    @classmethod
    def _build_id_names_parameters_schema(cls, id_names, schema, in_):
        parameters = []
        for id_name in id_names:
            parameter = deepcopy(schema['properties'][id_name])
            parameter.update({
                'name': id_name,
                'in': in_
            })
            if in_ == 'path':
                parameter['required'] = True
            parameters.append(parameter)

        return parameters

    @classmethod
    def _get_items_collections_metaclass(cls):
        return ItemsCollectionsModelBaseMeta

    @classmethod
    def insert(cls, session, objs, commit=True, todict=True, **kwargs):
        objs = type(cls).insert(cls, session, objs, commit=commit, todict=todict, **kwargs)
        if cls.__build_items_models__:
            cls.associate_items(objs)

        return objs

    @classmethod
    def update(cls, session, objs, commit=True, todict=True, ids=None, **kwargs):
        old_items_types = cls.get(session, ids=ids)
        objs = type(cls).update(cls, session, objs, commit=commit, todict=todict, ids=ids, **kwargs)
        cls.reassociate_items(old_items_types, objs)
        return objs

    @classmethod
    def reassociate_items(cls, old_items_types, new_items_types):
        new_names = [item_type['name'] for item_type in new_items_types]
        old_items_types = \
            [item_type for item_type in old_items_types if item_type['name'] not in new_names]

        for item_type in old_items_types:
            cls._disassociate_item(item_type)

        for item_type in new_items_types:
            cls._disassociate_item(item_type)
            cls._build_items_collections_model(item_type)

    @classmethod
    def _disassociate_item(cls, item_type):
        name = item_type['name']
        if name in cls.__api__.models:
            cls.__api__.disassociate_model(cls.__api__.models[name])

    @classmethod
    def delete(cls, session, ids, commit=True, **kwargs):
        items_types = cls.get(session, ids=ids)
        type(cls).delete(cls, session, ids, commit=commit, **kwargs)
        [cls._disassociate_item(item_type) for item_type in items_types]

    @staticmethod
    def get_module_path():
        return get_dir_path(__file__)


class ItemsTypesModelFiltersUpdaterBase(ItemsTypesModelBase):

    @classmethod
    def _build_items_collections_schema(cls, key, schema, id_names):
        update_filters_uri = '/{}/update_filters'.format(key)
        schema = ItemsTypesModelBase._build_items_collections_schema(key, schema, id_names)
        schema[update_filters_uri] = {
            'parameters': [{
                'name': 'Authorization',
                'in': 'header',
                'required': True,
                'type': 'string'
            },{
                'name': 'store_id',
                'in': 'query',
                'required': True,
                'type': 'integer'
            }],
            'post': {
                'operationId': 'post_job',
                'responses': {'201': {'description': 'Executing'}}
            },
            'get': {
                'parameters': [{
                    'name': 'hash',
                    'in': 'query',
                    'required': True,
                    'type': 'string'
                }],
                'operationId': 'get_job',
                'responses': {'200': {'description': 'Got'}}
            }
        }
        return schema

    @classmethod
    def _get_items_collections_metaclass(cls):
        return ItemsCollectionsModelFiltersUpdaterBaseMeta


class ItemsModelBaseMeta(ModelRedisMeta):

    def __init__(cls, name, bases, attrs):
        ModelRedisMeta.__init__(cls, name, bases, attrs)
        cls.index = None

    def get(cls, session, ids=None, limit=None, offset=None, **kwargs):
        items_per_page, page = kwargs.get('items_per_page', 1000), kwargs.get('page', 1)
        limit = items_per_page * page
        offset = items_per_page * (page-1)
        return ModelRedisMeta.get(cls, session, ids=ids, limit=limit, offset=offset, **kwargs)

    def get_all(cls, session, **kwargs):
        return ModelRedisMeta.get(cls, session, **kwargs)


class ItemsCollectionsModelBaseMeta(ModelRedisMeta):

    def insert(cls, session, objs, **kwargs):
        items_model = cls._get_model(kwargs)
        ret_values = items_model.insert(session, objs, **kwargs)
        cls._set_stock_filter(session, items_model)
        return ret_values

    def _get_model(cls, kwargs):
        store_id = kwargs.pop('store_id')
        items_model = cls.__models__.get(store_id)
        if items_model is None:
            raise HTTPNotFound()
        return items_model

    def _set_stock_filter(cls, session, items_model):
        items_keys = set(session.redis_bind.hkeys(items_model.__key__))
        items_indices_map = ItemsIndicesMap(items_model).get_all(session)
        items_indices_keys = set(items_indices_map.keys())
        remaining_keys = items_indices_keys.intersection(items_keys)
        old_keys = items_indices_keys.difference(items_keys)

        items = []
        cls._set_stock_item(remaining_keys, items_model, items_indices_map, True, items)
        cls._set_stock_item(old_keys, items_model, items_indices_map, False, items)

        stock_filter = BooleanFilterBy(items_model, 'stock')
        stock_filter.update(session, items)

    def _set_stock_item(cls, keys, items_model, items_indices_map, value, items):
        for k in keys:
            item = {k_: v_ for k_, v_ in zip(items_model.__id_names__, eval(k))}
            item.update({'stock': value, 'index': int(items_indices_map[k])})
            items.append(item)

    def update(cls, session, objs, ids=None, **kwargs):
        items_model = cls._get_model(kwargs)
        ret_values = items_model.update(session, objs, ids=ids, **kwargs)
        cls._set_stock_filter(session, items_model)
        return ret_values

    def get(cls, session, ids=None, limit=None, offset=None, **kwargs):
        items_model = cls._get_model(kwargs)
        return items_model.get(session, ids=ids, limit=limit, offset=offset, **kwargs)


class ItemsCollectionsModelFiltersUpdaterBaseMeta(ItemsCollectionsModelBaseMeta):

    def _run_job(cls, req, resp):
        job_session = req.context['job_session']
        query_string = req.context['parameters']['query_string']
        store_id = query_string['store_id']
        items_model = cls._get_model(query_string)
        items_indices_map = ItemsIndicesMap(items_model)
        items_indices_map_ret = items_indices_map.update(job_session)

        filters_factory = FiltersFactory()
        enabled_filters = cls._get_enabled_filters(job_session, store_id)
        filters_ret = dict()
        items_indices_map = items_indices_map.get_all(job_session)
        items = cls._get_items_with_indices_and_stock(job_session, items_model, items_indices_map)

        stock_filter = BooleanFilterBy(items_model, 'stock')
        stock_filter.update(job_session, items)

        for slot_var, schema in enabled_filters:
            filter_ = filters_factory.make(items_model, slot_var, schema, slot_var['skip_values'])
            filters_ret[filter_.name] = filter_.update(job_session, items)

        return {'items_indices_map': items_indices_map_ret, 'filters': filters_ret}

    def _get_enabled_filters(cls, session, store_id):
        slots_model = cls.__all_models__['slots']
        slots = slots_model.get(session, **{'store_id': store_id})
        filters_variables = []

        # used to aggregate filters inclusive and exclusive with same property
        filters_names_set = set()

        for slot in slots:
            if cls.__item_type__['id'] == slot['engine']['item_type_id']:
                for slot_var in slot['slot_variables']:
                    if slot_var['is_filter']:
                        filter_name = slot_var['inside_engine_name']
                        if filter_name not in filters_names_set:
                            schema = cls.__item_type__['schema']['properties'][filter_name]
                            filters_variables.append((slot_var, schema))
                            filters_names_set.add(filter_name)

        return filters_variables

    def _get_items_with_indices_and_stock(cls, job_session, items_model, items_indices_map):
        items = []
        page = 1
        items_part = items_model.get(job_session, page=page, items_per_page=100000)

        while items_part:
            items.extend(items_part)
            page += 1
            items_part = items_model.get(job_session, page=page, items_per_page=100000)

        for item in items:
            index = items_indices_map.get(item)
            if index is not None:
                item['index'] = index
                item['stock'] = True

        return items


def build_items_types_stores_table(metadata, **kwargs):
    return sa.Table(
        'items_types_stores', metadata,
        sa.Column('item_type_id', sa.Integer, sa.ForeignKey('items_types.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('store_id', sa.Integer, sa.ForeignKey('stores.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.cores.base import EngineCore, EngineError, RedisObjectBase
from falconswagger.utils import get_model_schema
from falconswagger.json_builder import JsonBuilder
from concurrent.futures import ThreadPoolExecutor
import numpy as np
import os.path
import zlib


class TopSellerEngineCore(EngineCore):
    __configuration_schema__ = get_model_schema(__file__)

    def export_objects(self, session, items_indices_map):
        readers = self._build_csv_readers('top_seller')
        items_indices_map_dict = self._get_items_indices_map_dict(items_indices_map, session)

        top_seller = TopSellerRedisObject(self)
        top_seller.update(readers, session, items_indices_map_dict)

        result = sorted(
            enumerate(top_seller.numpy_array), key=(lambda x: (x[1], x[0])), reverse=True)
        indices_items_map = items_indices_map.get_indices_items_map(session)
        return [{self._format_output(indices_items_map, r): int(r[1])} for r in result]

    def _format_output(self, indices_items_map, r):
        return ' | '.join([str(i) for i in eval(indices_items_map[r[0]])])

    def _build_rec_vector(self, session, **variables):
        return TopSellerRedisObject(self).get_numpy_array(session)
        rec_vector = session.redis_bind.get(redis_key)
        if rec_vector:
            return np.fromstring(zlib.decompress(rec_vector), dtype=np.int32)


class TopSellerRedisObject(RedisObjectBase):

    def update(self, readers, session, items_indices_map_dict):
        self._build_top_seller_vector(readers, session, items_indices_map_dict)
        session.redis_bind.set(
            self._redis_key,
            zlib.compress(self.numpy_array.tobytes())
        )

    def _build_top_seller_vector(self, readers, session, items_indices_map_dict):
        error_message = "No data found for engine '{}'".format(self._engine_core.engine['name'])
        if not len(readers):
            raise EngineError(error_message)

        indices_values_map = dict()

        for reader in readers:
            self._set_indices_values_map(indices_values_map, items_indices_map_dict, reader)

        if not indices_values_map:
            raise EngineError(error_message)

        vector = np.zeros(max(indices_values_map.keys())+1, dtype=np.int32)
        indices = np.array(list(indices_values_map.keys()), dtype=np.int32)
        vector[indices] = np.array(list(indices_values_map.values()), dtype=np.int32)
        self.numpy_array = vector

    def _set_indices_values_map(self, indices_values_map, items_indices_map_dict, reader):
        for line in reader:
            value = line.pop('value')
            self._set_item_values(line)
            index = items_indices_map_dict.get(line)
            if index is not None:
                indices_values_map[int(index)] = int(value)

    def get_numpy_array(self, session):
        rec_vector = session.redis_bind.get(self._redis_key)
        if rec_vector:
            return np.fromstring(zlib.decompress(rec_vector), dtype=np.int32)



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from falconswagger.models.orm.session import Session
from unittest import mock
from sqlalchemy import create_engine

import pytest
import pymysql



@pytest.fixture
def redis():
    return mock.MagicMock()


@pytest.fixture(scope='session')
def engine(variables):
    if variables['database']['password']:
        url = 'mysql+pymysql://{user}:{password}'\
            '@{host}:{port}/{database}'.format(**variables['database'])
    else:
        variables['database'].pop('password')
        url = 'mysql+pymysql://{user}'\
            '@{host}:{port}/{database}'.format(**variables['database'])
        variables['database']['password'] = None

    return create_engine(url)


@pytest.fixture
def session(model_base, request, variables, redis, engine):
    conn = pymysql.connect(
        user=variables['database']['user'], password=variables['database']['password'])

    with conn.cursor() as cursor:
        try:
            cursor.execute('drop database {};'.format(variables['database']['database']))
        except:
            pass

        cursor.execute('create database {};'.format(variables['database']['database']))
    conn.commit()
    conn.close()

    model_base.metadata.bind = engine
    model_base.metadata.create_all()

    session = Session(bind=engine, redis_bind=redis)

    def tear_down():
        session.close()

    request.addfinalizer(tear_down)
    return session



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.users.models import (GrantsModelBase, URIsModelBase, MethodsModelBase,
    UsersModelBase, build_users_grants_table, build_users_stores_table)
from myreco.stores.model import StoresModelBase
from myreco.variables.model import VariablesModelBase
from myreco.placements.models import (PlacementsModelBase, VariationsModelBase,
    ABTestUsersModelBase, build_variations_slots_table)
from myreco.slots.models import (SlotsVariablesModelBase,
    SlotsModelBase, build_slots_fallbacks_table)
from myreco.engines.models import EnginesModelBase, EnginesCoresModelBase
from myreco.engines.cores.base import EngineCore, AbstractDataImporter
from myreco.engines.cores.utils import build_engine_data_path
from myreco.items_types.models import ItemsTypesModelBase
from myreco.factory import ModelsFactory
from unittest.mock import MagicMock
from os import makedirs
from csv import DictWriter
import os.path
import gzip


table_args = {'mysql_engine':'innodb'}
factory = ModelsFactory('myreco', commons_models_attributes={'__table_args__': table_args},
    					commons_tables_attributes=table_args)
models = factory.make_all_models()

GrantsModel = models['grants']

URIsModel = models['uris']

MethodsModel = models['methods']

UsersModel = models['users']

StoresModel = models['stores']

VariablesModel = models['variables']

PlacementsModel = models['placements']

VariationsModel = models['variations']

ABTestUsersModel = models['ab_test_users']

SlotsVariablesModel = models['slots_variables']

SlotsModel = models['slots']

EnginesModel = models['engines']

EnginesCoresModel = models['engines_cores']

ItemsTypesModel = models['items_types']
    
SQLAlchemyRedisModelBase = factory.base_model

DataImporter = MagicMock()


class TestDataImporter(AbstractDataImporter):

        def get_data(self, items_indices_map, session):
            data_path = build_engine_data_path(self._engine)
            if not os.path.isdir(data_path):
                makedirs(data_path)

            filename_prefix = 'top_seller'
            data = [{'item_id': 2, 'sku': 'test2', 'value': 1},
                    {'item_id': 1, 'sku': 'test1', 'value': 3},
                    {'item_id': 3, 'sku': 'test3', 'value': 2}]
            file_ = gzip.open(os.path.join(data_path, filename_prefix) + '-000000001.gz', 'wt')
            writer = DictWriter(file_, ['sku', 'value', 'item_id'], delimiter='#')
            writer.writeheader()
            writer.writerows(data)
            file_.close()
            return {'lines_count': len(data)}


class TestEngine(EngineCore):
    __configuration_schema__ = {
        "type": "object",
        "required": ["item_id_name", "aggregators_ids_name"],
        "properties": {
            "item_id_name": {"type": "string"},
            "aggregators_ids_name": {"type": "string"}
        }
    }

    def get_variables(self):
        item_id_name = self.engine['configuration']['item_id_name']
        aggregators_ids_name = self.engine['configuration']['aggregators_ids_name']
        item_type_schema_props = self.engine['item_type']['schema']['properties']
        return [{
            'name': item_id_name,
            'schema': item_type_schema_props[item_id_name]
        },{
            'name': aggregators_ids_name,
            'schema': item_type_schema_props[aggregators_ids_name]
        }]

    def _validate_config(self, engine):
        item_id_name = engine['configuration']['item_id_name']
        aggregators_ids_name = engine['configuration']['aggregators_ids_name']
        item_type_schema_props = engine['item_type']['schema']['properties']
        message = "Configuration key '{}' not in item_type schema"

        if item_id_name not in item_type_schema_props:
            raise ValidationError(message.format('item_id_name'),
                instance=engine['configuration'], schema=item_type_schema_props)

        elif aggregators_ids_name not in item_type_schema_props:
            raise ValidationError(message.format('aggregators_ids_name'),
                instance=engine['configuration'], schema=item_type_schema_props)

    def _build_rec_vector(self):
        pass

    get_recommendations = MagicMock()